package eu.jonahbauer.wizard.core.machine;

import eu.jonahbauer.wizard.core.machine.states.State;
import org.jetbrains.annotations.Blocking;
import org.jetbrains.annotations.NotNull;

import java.util.concurrent.*;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

public abstract class Context<S extends State<S,C>, C extends Context<S,C>> {
    protected final ScheduledExecutorService scheduler = Executors.newSingleThreadScheduledExecutor();
    protected S state;
    protected final ReentrantLock lock = new ReentrantLock();
    private final Condition finishCondition = lock.newCondition();
    private boolean finished;
    private Throwable exception;

    protected void start(@NotNull S state) {
        lock.lock();
        try {
            if (finished) throw new IllegalStateException("Context has already finished.");
            transition(null, state);
        } finally {
            lock.unlock();
        }
    }

    public void transition(S currentState, S newState) {
        lock.lock();
        try {
            if (state == currentState) {
                state = newState;
                if (currentState != null) //noinspection unchecked
                    currentState.onExit((C) this);
                onTransition(currentState, newState);
                if (newState != null) //noinspection unchecked
                    newState.onEnter((C) this);
            } else {
                throw new IllegalStateException("Current state does not match.");
            }
        } catch (Throwable t) {
            handleError(t);
        } finally {
            lock.unlock();
        }
    }

    public void finish() {
        finish(null);
    }

    public void finish(Throwable exception) {
        lock.lock();
        try {
            finished = true;
            this.exception = exception;
            finishCondition.signalAll();
            transition(state, null);
            scheduler.shutdown();
        } finally {
            lock.unlock();
        }
    }

    public void cancel() {
        finish(new CancellationException());
    }

    @SuppressWarnings("BooleanMethodIsAlwaysInverted")
    public boolean isDone() {
        return finished;
    }

    @Blocking
    public void await() throws InterruptedException, ExecutionException, CancellationException {
        lock.lock();
        try {
            while (!finished) {
                finishCondition.await();
            }
            if (exception != null) {
                if (exception instanceof CancellationException cancelled) {
                    throw cancelled;
                } else {
                    throw new ExecutionException(exception);
                }
            }
        } finally {
            lock.unlock();
        }
    }

    public void timeout(@NotNull S currentState, long delay) {
        scheduler.schedule(() -> {
            lock.lock();
            try {
                if (state == currentState) {
                    //noinspection unchecked
                    state.onTimeout((C) this);
                }
            } catch (Throwable t) {
                handleError(t);
            } finally {
                lock.unlock();
            }
        }, delay, TimeUnit.MILLISECONDS);
    }

    protected void handleError(Throwable t) {
        lock.lock();
        try {
            if (!isDone()) {
                finish(t);
                t.printStackTrace();
            }
        } finally {
            lock.unlock();
        }
    }

    protected void onTransition(S from, S to) {}
}
