package eu.jonahbauer.wizard.core.machine.states.round;

import eu.jonahbauer.wizard.core.machine.states.GameData;
import eu.jonahbauer.wizard.core.machine.Game;
import eu.jonahbauer.wizard.core.machine.states.InvalidDataException;
import eu.jonahbauer.wizard.core.machine.states.game.Finishing;
import eu.jonahbauer.wizard.core.messages.observer.ScoreMessage;
import org.jetbrains.annotations.VisibleForTesting;

import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import static eu.jonahbauer.wizard.core.machine.states.GameData.*;

public final class FinishingRound extends RoundState {

    public FinishingRound(GameData data) {
        super(data.requireEach(PLAYERS, PREDICTIONS).require(TRICKS, SCORE));
        checkData(data);
    }

    @Override
    public void onEnter(Game game) {
        var points = getPoints();
        game.notify(new ScoreMessage(points));

        var score = new HashMap<>(get(SCORE));
        points.forEach((uuid, amount) -> score.merge(uuid, amount, Integer::sum));

        GameData data = getData().with(SCORE, Map.copyOf(score));

        if (60 / getPlayerCount() == get(ROUND) + 1) {
            transition(game, new Finishing(data));
        } else {
            transition(game, new StartingRound(data.with(ROUND, get(ROUND) + 1)));
        }
    }

    @VisibleForTesting
    Map<UUID, Integer> getPoints() {
        var points = new HashMap<UUID, Integer>();
        for (UUID player : get(PLAYERS)) {
            int prediction = get(PREDICTIONS).get(player);
            int tricks = get(TRICKS).getOrDefault(player, 0);

            if (tricks == prediction) {
                points.put(player, 20 + 10 * tricks);
            } else {
                points.put(player, -10 * Math.abs(tricks - prediction));
            }
        }
        return points;
    }

    private static void checkData(GameData data) {
        // the number of tricks played should be equal to the number of tricks in total
        int tricks = data.get(TRICKS).values().stream().mapToInt(i -> i).sum();
        if (tricks != data.get(ROUND) + 1) {
            throw new InvalidDataException("Unexpected number of tricks in round " + data.get(ROUND) + ": " + tricks + ".");
        }
    }
}
