package eu.jonahbauer.wizard.core.machine.states.trick;

import eu.jonahbauer.wizard.core.machine.states.GameData;
import eu.jonahbauer.wizard.core.machine.Game;
import eu.jonahbauer.wizard.core.machine.states.InvalidDataException;
import eu.jonahbauer.wizard.core.machine.states.round.FinishingRound;
import eu.jonahbauer.wizard.core.messages.observer.TrickMessage;
import eu.jonahbauer.wizard.core.model.Card;
import eu.jonahbauer.wizard.core.util.Pair;
import org.jetbrains.annotations.VisibleForTesting;

import java.util.*;

import static eu.jonahbauer.wizard.core.machine.states.GameData.*;

public final class FinishingTrick extends TrickState {

    public FinishingTrick(GameData data) {
        super(data.require(STACK));
        checkData(data);
    }

    @Override
    public void onEnter(Game game) {
        var winner = getWinner();
        game.notify(new TrickMessage(winner, get(STACK).stream().map(Pair::second).toList()));

        var tricks = new HashMap<>(get(TRICKS));
        tricks.merge(winner, 1, Integer::sum);

        GameData data = getData().with(TRICKS, Map.copyOf(tricks));

        if (get(TRICK) < get(ROUND)) {
            transition(game, new StartingTrick(data.with(
                    CURRENT_PLAYER, winner,
                    TRICK, get(TRICK) + 1
            )));
        } else {
            transition(game, new FinishingRound(data));
        }
    }

    @VisibleForTesting
    UUID getWinner() {
        var wizard = get(STACK).stream()
                               .filter(pair -> pair.second().getSuit() == Card.Suit.WIZARD)
                               .findFirst()
                               .orElse(null);
        if (wizard != null) {
            return wizard.first();
        }

        if (get(STACK).stream().allMatch(pair -> pair.second().getSuit() == Card.Suit.JESTER)) {
            return get(STACK).get(0).first();
        }

        var trumpSuit = get(TRUMP_SUIT);
        var suit = getTrickSuit();
        return get(STACK).stream()
                         .max(
                                 Comparator.<Pair<UUID, Card>>comparingInt(pair -> pair.second().getSuit() == trumpSuit ? 1 : 0)
                                           .thenComparing(pair -> pair.second().getSuit() == suit ? 1 : 0)
                                           .thenComparing(pair -> pair.second().getValue())
                         )
                         .orElseThrow(AssertionError::new)
                         .first();
    }

    private static void checkData(GameData data) {
        // every player has played exactly one card
        int stackSize = data.get(STACK).size();
        int playerCount = data.get(PLAYERS).size();
        if (stackSize != playerCount) {
            throw new InvalidDataException("Unexpected stack size: " + stackSize + " (expected " + playerCount + ").");
        }

        // every player has round - trick cards left on his hand
        int expectedSize = data.get(ROUND) - data.get(TRICK);
        for (List<Card> hand : data.get(HANDS).values()) {
            if (hand.size() != expectedSize) {
                throw new AssertionError("Unexpected hand size: " + hand.size() + " (expected " + expectedSize + ").");
            }
        }
    }
}
