package eu.jonahbauer.wizard.core.machine.states.trick;

import eu.jonahbauer.wizard.core.machine.states.GameData;
import eu.jonahbauer.wizard.core.machine.Game;
import eu.jonahbauer.wizard.core.machine.states.InvalidDataException;
import eu.jonahbauer.wizard.core.messages.observer.CardMessage;
import eu.jonahbauer.wizard.core.messages.observer.UserInputMessage;
import eu.jonahbauer.wizard.core.messages.player.PlayCardMessage;
import eu.jonahbauer.wizard.core.messages.player.PlayerMessage;
import eu.jonahbauer.wizard.core.model.Card;
import eu.jonahbauer.wizard.core.util.Pair;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.VisibleForTesting;

import java.util.*;

import static eu.jonahbauer.wizard.core.machine.states.GameData.*;
import static eu.jonahbauer.wizard.core.messages.observer.UserInputMessage.Action.PLAY_CARD;

public final class PlayingCard extends TrickState {

    public PlayingCard(GameData data) {
        super(data.require(STACK));
        checkData(data);
    }

    @Override
    public void onEnter(Game game) {
        game.notify(new UserInputMessage(get(CURRENT_PLAYER), PLAY_CARD, getTimeout(game, true)));
        timeout(game);
    }

    @Override
    public void onMessage(Game game, UUID player, PlayerMessage message) {
        if (get(CURRENT_PLAYER).equals(player) && message instanceof PlayCardMessage cardMessage) {
            checkCard(cardMessage.getCard());
            transition(game, cardMessage.getCard());
        } else {
            super.onMessage(game, player, message);
        }
    }

    @Override
    public void onTimeout(Game game) {
        var hand = get(HANDS).get(get(CURRENT_PLAYER));
        var card = hand.stream().filter(c -> {
            try {
                checkCard(c);
                return true;
            } catch (IllegalArgumentException e) {
                return false;
            }
        }).findAny().orElseThrow(() -> new AssertionError("Cannot play any card."));
        transition(game, card);
    }

    @VisibleForTesting
    void checkCard(Card card) {
        var hand = get(HANDS).get(get(CURRENT_PLAYER));
        if (!hand.contains(card)) {
            throw new IllegalArgumentException("You do not have this card on your hand.");
        }

        Card.Suit suit = getTrickSuit();
        if (card.getSuit().isColor() && suit != null && card.getSuit() != suit) {
            if (hand.stream().anyMatch(c -> c.getSuit() == suit)) {
                throw new IllegalArgumentException("Must follow suit.");
            }
        }
    }

    private void transition(Game game, @NotNull Card card) {
        var currentPlayer = get(CURRENT_PLAYER);
        game.notify(new CardMessage(currentPlayer, card));

        // add card to stack
        var stack = get(STACK);
        var mutableStack = new ArrayList<Pair<UUID, Card>>(stack.size() + 1);
        mutableStack.addAll(stack);
        mutableStack.add(Pair.of(currentPlayer, card));

        // remove card from hand
        var hands = get(HANDS);
        var mutableHands = new HashMap<>(hands);
        var hand = hands.get(currentPlayer);
        var mutableHand = new ArrayList<>(hand);
        mutableHand.remove(card);
        mutableHands.put(currentPlayer, List.copyOf(mutableHand));

        GameData data = getData().with(
                STACK, List.copyOf(mutableStack),
                HANDS, Map.copyOf(mutableHands)
        );

        var summary = data.get(HANDS).values().stream()
                          .mapToInt(Collection::size)
                          .summaryStatistics();

        if (summary.getMax() == summary.getMin()) { // everybody has the same amount of cards
            transition(game, new FinishingTrick(data));
        } else {
            transition(game, new PlayingCard(data.with(CURRENT_PLAYER, getNextPlayer())));
        }

    }

    private static void checkData(GameData data) {
        int expectedSize = data.get(ROUND) - data.get(TRICK);
        for (List<Card> hand : data.get(HANDS).values()) {
            if (hand.size() != expectedSize && hand.size() != expectedSize + 1) {
                throw new InvalidDataException("Unexpected hand size: " + hand.size() + " (expected " + expectedSize + " or " + (expectedSize + 1) + ").");
            }
        }
        int size = data.get(HANDS).get(data.get(CURRENT_PLAYER)).size();
        if (size != expectedSize + 1) {
            throw new InvalidDataException("Unexpected current player's hand size: " + size + " (expected " + (expectedSize + 1) + ")");
        }
    }
}
